//
// This file is part of Nolimips, a MIPS simulator with unlimited registers
// Copyright (C) 2003, 2004, 2006 Benoit Perrot <benoit@lrde.epita.fr>
//
// Nolimips is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// Nolimips is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
//
/** \file vm/cpu.hh
    \brief Declare Central Processor Unit class. */
#ifndef VM_CPU_HH
# define VM_CPU_HH

# include <stdint.h>
# include <vector>

# include "misc/table.hh"

# include "inst/inst.hh"
# include "inst/visitor.hh"

# include "vm/fwd.hh"

namespace vm
{

  typedef int32_t	register_type;
  typedef uint32_t	uregister_type;

  /// Central Processor Unit abstraction
  class Cpu:
    protected inst::ConstVisitor
  {
  public:
    /// General purpose registers indices
    enum kind_type
      {
	zero = 0,
	at,
	v0, v1,
	a0, a1, a2, a3,
	t0, t1, t2, t3, t4, t5, t6, t7,
	s0, s1, s2, s3, s4, s5, s6, s7,
	t8, t9,
	k0, k1,
	gp,
	sp,
	fp,
	ra = 31
      };

    /** \name Constructor and destructor
	\{ */
  public:
    /// Construct a CPU.
    Cpu(Mmu &mmu, Cp0 &cp0);
    /// Destroy a CPU.
    virtual ~Cpu();
    /** \} */

  public:
    /// Reset the CPU (set general purpose registers to zero, etc.)
    void reset();

    /// Print registers' content
    void write(std::ostream &os) const;

    /** \name General purpose registers accessors.
	\{ */
  public:
    /// Set the general purpose register \a k to \a r.
    void set_register(kind_type k, register_type r);
    /// Return the general purpose register \a k.
    register_type get_register(kind_type k) const;
    /** \} */

    /** \name Unlimited registers accessors.
	\{ */
  public:
    /// Set the unlimited register indexed by \a i to \a r.
    void set_unlimited(int i, register_type r);
    /// Return true if the unlimited register \a i exists.
    bool has_unlimited(int i) const;
    /// Return the unlimited register indexed by \a i.
    register_type get_unlimited(int i) const;
    /** \} */

    /** \name Registers unified accessors.
	\{ */
  public:
    /// Set the register identified by \a reg to \a r.
    void set_register(const inst::Register &reg, register_type r);
    /// Return the register identified by \a reg.
    register_type get_register(const inst::Register &reg) const;
    /** \} */

    /** \name Special registers accessors.
	\{ */
  public:
    /// Set the HI register to \a r.
    void set_hi(register_type r);
    /// Return the HI register.
    register_type get_hi() const;

    /// Set the LO register to \a r.
    void set_lo(register_type r);
    /// Return the LO register.
    register_type get_lo() const;

    /// Set the PC register to \a r.
    void set_pc(register_type r);
    /// Return the PC register.
    register_type get_pc() const;
    /** \} */

  public:
    Mmu &get_mmu();
    Cp0 &get_cp0();

  public:
    unsigned get_instruction_counter(inst::Inst::format_type format) const;

    const std::vector<register_type> &get_call_stack() const;
    void set_check_callee_save(bool check_callee_save_p);

  public:
    /// Set trace flag
    void set_trace(bool trace_p);

    /// Make one execution step
    void step();

  protected:
    virtual void	visit(const inst::Add& add);
    virtual void	visit(const inst::Addi& addi);
    virtual void	visit(const inst::Addu& addu);
    virtual void	visit(const inst::Addiu& addiu);
    virtual void	visit(const inst::Sub& sub);
    virtual void	visit(const inst::Subu& subu);

    virtual void	visit(const inst::Sll& sll);
    virtual void	visit(const inst::Sllv& slv);
    virtual void	visit(const inst::Sra& sra);
    virtual void	visit(const inst::Srav& srav);
    virtual void	visit(const inst::Srl& srl);
    virtual void	visit(const inst::Srlv& srlv);

    virtual void	visit(const inst::Mul& mul);
    virtual void	visit(const inst::Div& div);
    virtual void	visit(const inst::Divu& divu);

    virtual void	visit(const inst::Lb& lb);
    virtual void	visit(const inst::Lbu& Lbu);
    virtual void	visit(const inst::Lw& lw);
    virtual void	visit(const inst::Lui& lui);
    virtual void	visit(const inst::Sb& sb);
    virtual void	visit(const inst::Sw& sw);

    virtual void	visit(const inst::And& _and);
    virtual void	visit(const inst::Andi& andi);
    virtual void	visit(const inst::Or& _or);
    virtual void	visit(const inst::Ori& ori);
    virtual void	visit(const inst::Nor& nor);
    virtual void	visit(const inst::Xor& _xor);
    virtual void	visit(const inst::Xori& xori);

    virtual void	visit(const inst::Slt& slt);
    virtual void	visit(const inst::Slti& slti);
    virtual void	visit(const inst::Sltu&);
    virtual void	visit(const inst::Sltiu& sltiu);

    virtual void	visit(const inst::Jmp& jmp);
    virtual void	visit(const inst::Jr& jr);
    virtual void	visit(const inst::Jal& jal);
    virtual void	visit(const inst::Jalr& jalr);

    virtual void	visit(const inst::Beq& beq);
    virtual void	visit(const inst::Bne& bne);

    virtual void	visit(const inst::Bgez& bgez);
    virtual void	visit(const inst::Bgezal& bgezal);
    virtual void	visit(const inst::Bgtz& bgtz);

    virtual void	visit(const inst::Blez& blez);
    virtual void	visit(const inst::Bltz& bltz);
    virtual void	visit(const inst::Bltzal& bltzal);

    virtual void	visit(const inst::Mfhi& mfhi);
    virtual void	visit(const inst::Mflo& mflo);
    virtual void	visit(const inst::Mthi& mthi);
    virtual void	visit(const inst::Mtlo& mtlo);
    virtual void	visit(const inst::Mfc0& mfc0);
    virtual void	visit(const inst::Mtc0& mtc0);

    virtual void	visit(const inst::Syscall& sycall);

    void call();
    void ret();

  protected:
    /// Memory management unit link.
    Mmu &mmu_;
    /// Control coprocessor link.
    Cp0 &cp0_;

    /// General purpose registers.
    register_type GPR_[32];
    /// Special HIgh register.
    register_type hi_;
    /// Special LOw register.
    register_type lo_;
    /// Special Program Counter.
    register_type pc_;

    /// Unlimited registers.
    misc::Table<int, register_type> unlimited_;

  protected:
    std::vector<register_type> call_stack_;
    bool check_callee_save_p_;

  protected:
    bool trace_p_;

    enum pipeline_stage_type
      {
	i_stage = 0,	///< Instruction fetch stage
	d_stage,	///< Instruction dispatch stage
	r_stage,	///< Regoster file read stage
	e_stage,	///< Execution stage
	m_stage,	///< Memory access stage
	w_stage		///< Writeback stage
      };
    inst::Inst *bubble_;

    /// Instruction pipeline
    const inst::Inst *pipeline_[6];

    /// Instruction counters
    unsigned counters_[3];
  };

  std::ostream &
  operator<<(std::ostream &ostr, const Cpu &cpu);

} // namespace vm

# include "vm/cpu.hxx"

#endif // !VM_CPU_HH
